------------------------------------------------------------------------------
-- Attribute handler, this contains functions made to mess with attributes	--
--				it also contains stance switching handler code				--
------------------------------------------------------------------------------
local Attributes =  FlexBar2:NewModule("Attributes");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings
Attributes.Version = 1;

-- Get an instance of this lovely SecureStateDriverFrame for use with stance, shift & stealth changes
local SecureStateDriverFrame = CreateFrame("Frame", nil, UIParent, "SecureStateDriverTemplate");

-- Function called after FlexBar2's OnInitialize by Ace2
function Attributes:Activate()
	SecureStateDriverFrame:SetAttribute("statemap-form", "$input");
	self:SetShapeshiftStates();
end

local Active_statebutton, Inactive_statebutton;
-- declare all shapeshift stuff, quite complicated func
function Attributes:SetShapeshiftStates()
	local class = UnitClass("player");
	local NumShapeshifts = GetNumShapeshiftForms();
	-- Add a shapeshift amount if the player is a druid, this is to add prowled cat in as it aint a different state then cat by default
	if(class == "DRUID") then
		NumShapeshifts = NumShapeshifts + 1;
	end
	-- what the statebutton is set to when the button should listen to state changes, typically this is 0:S0;1:S1;2:S2;...
	Active_statebutton = "";
	-- What the statebutton is set to when it is NOT listening to state changes, typically this is 0:S0;1:S0;2:S0;...
	Inactive_statebutton = "";
	-- 'magic' code passed to the state driver to make our state changes happen, typically this is something like [stance:0]0;[stance1:1,stealth:1]1;... depending on your class
	local StateDriverMagic = "";
	-- Loop through every shapeshift and populate Active_statebutton & Inactive_statebutton
	for State = 0, NumShapeshifts do
		Active_statebutton = Active_statebutton .. State .. ":S" .. State .. ";";
		Inactive_statebutton = Inactive_statebutton .. State .. ":S0;";
	end
	-- Loop through the shapeshifts again in order to declare the StateDriverMagic var, note, we do nothing for state 0, its added as a fallback state at the end
	for State = 1, NumShapeshifts do
		-- cat & prowled cat managing code, the code is set for both state 2 & 3 because we are unsure wether state 2 or 3 is cat, as it depends wether the player has done the seal form quest, so set it for both to be sure
		if(class == "DRUID" and (State == 3 or State == 2)) then
			-- Add [stance:(stance),stealth:1]4;[stance:(stance),stealth:0](stance);
			StateDriverMagic = StateDriverMagic .. "[stance:" .. State .. ",stealth:1]4;[stance:" .. State .. ",stealth:0]" .. State .. ";";
			if(State == 3) then	State = 4; end
		else
			-- Add [stance:(stance)](stance);
			StateDriverMagic = StateDriverMagic .. "[stance:" .. State .. "]" .. State .. ";";
		end
	end
	-- Append 0 at the end of the state driver magicas fallback state
	StateDriverMagic = StateDriverMagic .. "0";
	RegisterStateDriver(SecureStateDriverFrame, "form", StateDriverMagic);
	-- Update all buttons after this has been done
	FlexBar2:Debug("state driver magic declared: ", StateDriverMagic);
	Attributes.StateDriverMagic = StateDriverMagic;
end

function Attributes:GetStateButtons()
	return Active_statebutton, Inactive_statebutton;
end

-- Declare the button mixin
Attributes.ButtonMixin = FlexBar2.OO.Mixin
{
	"LoadStateButton",
	"GetAttribute",
	"GetModifiedAttribute",
    "SetAttribute",
    "SetModifiedAttribute",
	"SetShapeshiftStates",
};

-- During initialization add this button to the state driver's list of buttons to manage
function Attributes.ButtonMixin:Initialize()
    SecureStateDriverFrame:SetAttribute("addchild", self.Frame);
end

-- Sets state changing to false by default and all attributes to nil
function Attributes.ButtonMixin:SetDefaults()
    self.InfoTable.Attributes = {};
	self.InfoTable.StateButton = false;
end

-- Load saved settings
function Attributes.ButtonMixin:LoadSettings()
    for Attribute, Value in pairs(self.InfoTable.Attributes) do
        self:SetAttribute(Attribute, Value, false);
    end
	self:LoadStateButton();
end

function Attributes.ButtonMixin:LoadStateButton()
	-- Set the state button attribute
	if(self.InfoTable.StateButton == true) then
		FlexBar2:Debug("setting state button attribute to ", Active_statebutton);
		self:SetAttribute("statebutton", Active_statebutton, false);
	else
		FlexBar2:Debug("setting state button attribute to ", Inactive_statebutton);
		self:SetAttribute("statebutton", Inactive_statebutton, false);
	end
	-- Reset the state
	SecureStateDriverFrame:SetAttribute("state", "0");
	SecureStateDriverFrame:SetAttribute("state", GetShapeshiftForm());
end

-- Gets an attribute on the button, usable both in & out of combat, but not very usefull in combat as SetAttribute is not avail
function Attributes.ButtonMixin:GetAttribute(Attribute)
	return self.Frame:GetAttribute(Attribute);
end

-- This will call :GetAttribute with the button prefix (ctrl-alt-shift-) before it and the button suffix (the current state) after it. if nothing is found with a certain prefix * is used as prefix
function Attributes.ButtonMixin:GetModifiedAttribute(Attribute)
    if(Attribute) then
        return self:GetAttribute(SecureButton_GetModifierPrefix() .. Attribute .. SecureButton_GetButtonSuffix(SecureStateChild_GetEffectiveButton(self.Frame))) or 
        self:GetAttribute("*" .. Attribute .. SecureButton_GetButtonSuffix(SecureStateChild_GetEffectiveButton(self.Frame)));
    end
end

-- Set an attribute on the frame, not usable in combat
function Attributes.ButtonMixin:SetAttribute(Attribute, Value, Save)
    self.Frame:SetAttribute(Attribute, Value);
	if(Save ~= false) then self.InfoTable.Attributes[Attribute] = Value; end
    FlexBar2:Debug("Setting Attribute [", Attribute, "] to ", Value);
end

-- Same as :GetAttribute except it sets the attribute instead of getting it :)
function Attributes.ButtonMixin:SetModifiedAttribute(Attribute, Value, Save)
	local Prefix = SecureButton_GetModifierPrefix();
	if(Prefix == "") then Prefix = "*"; end
	self:SetAttribute(Prefix .. Attribute .. SecureButton_GetButtonSuffix(SecureStateChild_GetEffectiveButton(self.Frame)), Value, Save); 
end

-- GUI -- 
local Waterfall = FlexBar2.Waterfall;
local StateButton;
function Attributes.ButtonWaterfall(Action, ButtonName, ButtonTable)
	if(Action == "edit") then
		-- 'State button' checkbox --
		StateButton = FlexBar2:GetSetting(ButtonTable, "StateButton");
		FlexBar2:Debug(StateButton);
		Waterfall:AddControl(
		"type","checkbox",
		"text","State button",
		"width",300,
		"causeRefresh", false,
		"getFunc" ,function() return StateButton; end,
		"setFunc", function(Switch) StateButton = Switch; end);
	elseif(Action == "save") then
		local ButtonList = FlexBar2.Buttons;
		local Button = ButtonList[ButtonTable];
		-- Save wether the button is a statebutton or not
		if(StateButton == true) then
			FlexBar2:ApplySetting(ButtonTable, "StateButton", true);
		else
			FlexBar2:ApplySetting(ButtonTable, "StateButton", false);
		end
		-- Force a state reset after applying the new settings to make sure the button actualy gets it
		Button:LoadStateButton();
	end
end

-- SlashCommands --
local SlashCommands = FlexBar2.SlashCommands;
local function SlashCommand_StateButton(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	local StateButton = Params.statebutton;
	local ButtonList = FlexBar2.Buttons;
	-- Force a state reset after applying the new settings to make sure the button actualy gets it
	if(StateButton == "true") then
		FlexBar2:ApplySetting(ButtonTable, "StateButton", true);
	else
		FlexBar2:ApplySetting(ButtonTable, "StateButton", false);
	end
	-- Loop through given button range
	for _, ButtonName in ipairs(ButtonTable) do
		local Button = ButtonList[ButtonName];
		FlexBar2:Debug("statebuttons: ", Active_statebutton, Inactive_statebutton);
		-- Reload the statebutton attribute
		Button:LoadStateButton();
	end
end

-- Localize this neat verify function for faster access
local VerifyExistance = FlexBar2.SlashCommands.VerifyExistance;
-- AddCommand format :AddCommand(Command, Function, Group, Help)
-- AddParam format :AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc) 
-- everything here should be self explainative knowing the syntax
SlashCommands:AddCommand("statebutton", SlashCommand_StateButton, "attributes", "/FlexBar2 statebutton button=<Button> statebutton=<true/false>");
	SlashCommands:AddParam("statebutton", "button", true, nil, nil, VerifyExistance);
	SlashCommands:AddParam("statebutton", "statebutton", true, {"true", "false"}, nil, nil);
